#ifndef __RELOC_H
#define __RELOC_H


#include <assert.h>

#define RELEASE "1.02"

/* Represent bytes as 8-bit values tagged with meta-information. */
/* List nodes (struct source) are arena-allocated; no need to track the references. */

struct source {
	struct source	*next;
	unsigned short int	offset;
	unsigned char		used;
};

typedef struct {
	unsigned char		value;
	struct source	*src;
} value_t;

struct core {
	value_t			memory[65536];
	unsigned char	read[65536];		// These (read/written) are booleans, but
	unsigned char	written[65536];		// kept separate for performance reasons.
};

struct progbyte {
	unsigned char			flags;
	unsigned char			zpaddr[32];	// one bit per address, little-endian
	struct constraintlist	*constr;
};

#define PBF_DONT_RELOC	0x01
#define PBF_RELOC	0x02
#define PBF_USED_IN_ZP	0x04
#define PBF_USED_IN_MSB	0x08

enum {
	RET_SUCCESS,
	RET_HEADER,
	RET_RSID,
	RET_MUS,
	RET_BASIC,
	RET_PSID,
	RET_PARAM,
	RET_IO,
	RET_CONSTR,
	RET_ZPFULL,
	RET_VERIFY,
	RET_PLAYADDR,
	RET_RANGE,
	RET_CYCLES,
};

enum {
	ERR_OK,
	ERR_BRK,
	ERR_INTERNAL,
	ERR_ILLEGAL,
	ERR_CYCLES,
};

extern char *emulate_err[];

extern struct progbyte *progbytes;
extern int progsize;

extern unsigned short int reloc_start, reloc_end;
extern int add_constraints;
extern int verbose;
extern int do_zp_reloc;

struct source *cons_src(unsigned short int car, struct source *cdr);
void check_reloc_range(unsigned short int addr, struct source *lsb1, struct source *lsb2, struct source *msb);
void reloc_alike(value_t v1, value_t v2);
void used_for_zp_addr(struct source *src1, struct source *src2, unsigned char zpaddr);
void init_progbytes(unsigned short int loadaddr, unsigned short int loadsize);
int trivially_inconsistent();
int solver();
void free_arena();
void gc_arena(struct core *core);
void reloc_map(struct core *oldcore, unsigned short int reloc_offs);
int emulate(struct core *core, unsigned short int start_addr, unsigned char acc, int max_cycles);
void finalise_constraints(struct core *core);
void prealloc_cons_cells();

#if 1
static __inline void dont_reloc_at(unsigned short int offset) {
	assert(offset != 1);
	assert(offset < progsize + 2);
	progbytes[offset].flags |= PBF_DONT_RELOC;
}

static __inline void dont_reloc(struct source *src) {
	if(add_constraints) {
		while(src) {
			dont_reloc_at(src->offset);
			assert(src);
			src = src->next;
		}
	}
}
#else
void dont_reloc_at(unsigned short int offset);
void dont_reloc(struct source *src);
#endif

#endif
